#!python
# Copyright (c) 2015-2015 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

from xpcom.components import interfaces as Ci
from xpcom.components import classes as Cc
from xpcom import components
import threading
import os.path
import time
import re
import apsw

import logging
import json

log = logging.getLogger("koScopeSections-Py")
#log.setLevel(10)

prefs = Cc["@activestate.com/koPrefService;1"].getService(Ci.koIPrefService).prefs
koDirSvc  = Cc["@activestate.com/koDirs;1"].getService()
symbolsPath = os.path.join(koDirSvc.userDataDir, "codeintel", "symbols.db")

class koScopeSections:
    
    _com_interfaces_ = [Ci.koIScopeSections]
    _reg_desc_ = "Commando - Sections Scope"
    _reg_clsid_ = "{686a592a-526d-4a50-96c6-25d5d2745158}"
    _reg_contractid_ = "@activestate.com/commando/koScopeSections;1"
    
    searches = {}
    indexing = False
    activeUuid = False
    
    @components.ProxyToMainThreadAsync
    def stopActiveSearches(self):
        for uuid in self.searches.keys():
            log.debug("Stopping search: " + uuid)
            self.searches[uuid].stop()
            del self.searches[uuid]

    def onSearch(self, query, uuid, cwd, filePath, language, onResults, onComplete):
        log.debug(uuid + " - Starting Search: " + query)
        
        self.activeUuid = uuid
        self.stopActiveSearches()
        
        query = query.strip().lower()
        
        self.searches[uuid] = Searcher(main=self, query=query, cwd=cwd, filePath=filePath,
                                       language=language, uuid=uuid, onResults=onResults,
                                       onComplete=onComplete)
        t = threading.Thread(target=self.searches[uuid].start, name="Sections Scope Search %s" % uuid)
        t.setDaemon(True)
        t.start()
        
    @components.ProxyToMainThreadAsync
    def callback(self, cb, value = "", code = 0):
        if not cb:
            return
        
        try:
            cb.callback(code, json.dumps(value))
        except Exception as e:
            #log.debug(value)
            log.exception(e)

class Searcher:
    
    def __init__(self, main, query, uuid, cwd, filePath, language, onResults, onComplete):
        log.debug(uuid + " - Initializing Searcher")
        
        self.main = main
        self.query = query
        self.queryWords = filter(None, re.split("\W+", query))
        self.uuid = uuid
        self.cwd = cwd
        self.filePath = filePath
        self.language = language
        self._onResults = onResults
        self._onComplete = onComplete
        
        self.resultsPending = []
        self.resultTimer = None
        self.completeTimer = None
        
        self._stop = False
        self.conn = False
        
    def start(self):
        while not os.path.isfile(symbolsPath):
            if self._stop:
                return
            
            time.sleep(1)
            
        self.conn = DB()
        
        num = 0   
        maxResults = prefs.getLong("commando_search_max_results")
        
        clause = ""
        if self.query:
            clause = "* NEAR ".join(self.queryWords).replace("'","\\'")
            clause = "AND search MATCH '%s*'" % clause
        
        # Get symbols in current file
        query = """
            SELECT * FROM symbols WHERE filepath='%s' %s
            """ % (self.filePath, clause)
            
        columns = None
        for row in self.conn.fetchAll(query):
            if self._stop:
                return
            
            num = num + 1
            if num == maxResults:
                self.stop()
                return
            
            self.process(row)
        
        # Still going? Include symbols that aren't in our current file
        query = """
            SELECT * FROM symbols WHERE filepath!='%s' AND filepath LIKE '%s%%' %s
            """ % (self.filePath, self.cwd, clause)
        
        columns = None
        for row in self.conn.fetchAll(query):
            if self._stop:
                return
            
            num = num + 1
            if num == maxResults:
                self.stop()
                return
            
            self.process(row)
        
        # All done
        self.stop()
        
    def process(self, data):
        name = data["symbol"]
        matchable = data["search"]
        
        weight = 0
        if data["filepath"] == self.filePath:
            weight = weight + 20
            
        if data["lang"] == self.language:
            weight = weight + 20
        
        if matchable == self.query:
            weight = weight + 20
        elif matchable.startswith(self.query):
            weight = weight + 15
            
        if name == self.query:
            weight = weight + 20
        elif name.startswith(self.query):
            weight = weight + 15
        
        # Prioritize results that match a larger part of the result
        if len(self.query) > 0:
            mx = len(matchable)
            diff = mx - len(self.query)
            weight = weight + (((mx - diff) * 20) / mx)
            
        nameFull = name
        if data["parent"]:
            nameFull = "%s > %s" % (name, data["parent"])
            
        self.returnResult({
            "id": "doc-%s" % data["search"],
            "name": nameFull,
            "description": "%s in %s:%s (%s)" % (data["type"], os.path.basename(data["filepath"]), data["lineno"], data["lang"]),
            "icon": "chrome://komodo/skin/images/codeintel/cb_%s.svg" % data["type"],
            "scope": "scope-sections",
            "weight": weight,
            "data": data
        })
        
    def stop(self):
        if (self._stop):
            return # Already stopped
        
        self._stop = True
            
        if self.conn and self.conn.open:
            try:
                self.conn.close()
            except:
                # This usually means the connection was made on another thread
                # that thread will close the connection when it notices that
                # self._stop was set
                pass 
            
        self.onComplete()
        
    def onComplete(self):
        self.completeTimer = threading.Timer(0.06, self.onCompleteTimer)
        self.completeTimer.start()
            
    def onCompleteTimer(self):
        self.completeTimer = None
        self.main.callback(self._onComplete)
                
    def returnResult(self, result):
        if self._stop:
            return
        
        self.resultsPending.append(result)

        if not self.resultTimer:
            self.resultTimer = threading.Timer(0.05, self._returnResults)
            self.resultTimer.start()
            
    def _returnResults(self):
        self.resultTimer = None
        
        results = self.resultsPending
        self.resultsPending = []
        
        if self.uuid != self.main.activeUuid:
            return
        
        log.debug(self.uuid + " - Returning " + str(len(results)) + " results")
        self.main.callback(self._onResults, results)
        
        
class DB:
    
    def __init__(self):
        self.conn = apsw.Connection(symbolsPath)
        self.cursor = self.conn.cursor()
        self.open = True
        
    def query(self, query, args = None):
        if args:
            return self.cursor.execute(query, args)
        else:
            return self.cursor.execute(query)
        
    def fetch(self, query, args = None):
        self.query(query, args)
        return self._row(self.cursor.fetchone())
    
    def fetchAll(self, query, args = None):
        for row in self.query(query, args):
            yield self._row(row)
            
    def close(self):
        self.conn.close()
        self.open = False
       
    def _row(self, row):
        if not row:
            return None
        
        d = {}
        for idx, col in enumerate(self.cursor.description):
            d[col[0]] = row[idx]
        
        return d
