#!/usr/bin/env python
import os
import re
import json
from xml.etree import ElementTree
from xpcom import components
from xpcom.server import UnwrapObject

import logging
log = logging.getLogger("sublime-converter")
#log.setLevel(logging.DEBUG)
# #from dbgp.client import brk
# #brk(host="localhost", port=10000)

"""This class is used to convert Sublime Text snippets to Komodo Snippets.
If more use cases become sensible then it can be extended to accomodate more."""

class Sublime():
    def __init__(self):
        self.komodoToolStr  = """
        {
          "version": "1.0.12", 
          "type": "snippet", 
          "name": "name", 
          "value": [], 
          "auto_abbreviation": "true",
          "language":"",
          "treat_as_ejs": "false"
        }
        """
        self.relativePath = ""
        self.conversionFolder  = ""
        self.conversionFolderName = "sublime-converted-snippets"
        self.rootDir  = "" #used to generate a relative path when importing folders
        self.disabledVarsInFile  = "" # Reset in convertSublimeSnippets for each file
                                 # set to true in _disableVarsInTabstops so file is
                                 # written to list of unsupported configs.
        self.unsupportedSnippets  ="unsupported-snippets.md"
        self.unsupportedSnippetsPath = ""
        self.unsupportedOrigContent  = """
### What's this file about?
Certain aspects of Sublime Text snippets aren't supported. This file (the one
you're reading) lists the files that were found to have these unsupported
features.

### What did we do?

We changed stuff from the snippet you knew.

#### Variables
We "disabled" the unsupported variables by removing `$` from them and appending
`_NOT_SUPPORTED` to it so it's easy to find.

#### Filename
Also note that due to how snippets are triggered in Komodo the filename of the
snippet has been converted to the trigger of the original. The trigger is still
the same.

### Effected Variables
The only variables we don't support are the `$PARAM#` variables (we don't support
this mechanism) and the the `$TM_SOFT_TABS` (we automatically insert what you
have set in your preferences).  See the [Sublime Snippet
docs](http://docs.sublimetext.info/en/latest/extensibility/snippets.html?highlight=snippet#environment-variables)
for more information.

### Trouble Makers
"""
        self.convCount=0
        self.badFileCount=0
    
    def convertSublimeSnippetFolder(self, srcPath):
        """Searches a folder for Sublime snippets, converts them and saves them
        to a conversion folder. When completed the conversion folder path is
        returned"""
        try:
            self.createConvFolder(srcPath)
            self.createUnsupportedFile()
            self.rootDir = srcPath
            os.path.walk(srcPath, self.convertSublimeSnippets, True)
            self.addToOutputFile("\n\nConverted %s files.\n%s bad files found." %(self.convCount, self.badFileCount))
            return self.conversionFolder
        except Exception as e:
            log.error("Folder Conversion Failed: %s",e)

    # Import a Sublime text snippet
    # 
    # Move description into an EJS comment in the snippet rather than throw it
    # away Trigger becomes name in toolbox.
    # Parse the content:
    #     insert description
    #     parse content and swap Sublime variables for Komodo shortcuts
    # 
    # Example snippet:
    # <!-- See http://www.sublimetext.com/docs/snippets for more information -->
    # <snippet>
    # 	<content><![CDATA[${1:FIELDNAME} = models.AutoField($2)]]></content>
    # 	<tabTrigger>mauto</tabTrigger>
    #     <scope>source.python</scope>
    #     <description>AutoField (mauto)</description>
    # </snippet>
    # 
    # Docs ref: http://docs.sublimetext.info/en/latest/extensibility/snippets.html
    def convertSublimeSnippets(self, folderConversion, parentPath, srcFiles):
        """Takes a list of paths of potential Sublime snippets and converts
        them then saves them to a conversion folder. Returns the conversion
        folder.
        This function can be called on it's own but is also called from
        convertSublimeSnippetFolder in a os.path.walk() call. Hence
        folderConversion.
        parentPath is used to maintain any folder structure that the original
        snippet repo may have had."""
        try:
            if not folderConversion:
                destPath = self.createConvFolder(parentPath)
                self.createUnsupportedFile()
            else:
                if parentPath == self.conversionFolder or ".git" in parentPath:
                    # don't process the conversion folder silly
                    # also don't process a git dir
                    return
                self.relativePath = os.path.relpath(parentPath, self.rootDir)
                destPath = os.path.join(self.conversionFolder, self.relativePath)
            if not os.path.exists(destPath):
                os.mkdir(destPath)
            for srcFile in srcFiles:
                # os.path.walk gives a list of files names.
                # The file browser that is called before convertSublimeSnippets
                # gives a list of file paths.
                if folderConversion:
                    srcFile = os.path.join(parentPath, srcFile)
                ext = os.path.splitext(srcFile)[1]
                if not ext == ".sublime-snippet":
                    log.warn("Skipping non Sublime tool: %s", srcFile)
                else:
                    data = self.convertSnippet(srcFile)
                    path = os.path.join(destPath, data["name"] + ".komodotool")
                    with open(path, "w") as destFile:
                        destFile.write(json.dumps(data))
                        self.convCount+=1
            if not folderConversion:
                self.addToOutputFile("\n\nConverted %s files.\n%s bad files found." %(self.convCount, self.badFileCount))
                return self.conversionFolder
        except Exception as e:
            log.error("File list conversion failed: %s", e)
    # Take a file path to a suposed snippet, convert it and dump it into the
    # destination folder
    def convertSnippet(self, srcFile):
        """Convert a Sublime snippet into a json object.
        Return the json object to be saved as a Komodo snippet"""
        self.disabledVarsInFile  = False
        xmlContent=""
        # Get the files contents into XML form
        with open(srcFile, "r") as fileContent:
            try:
                xmlContent = ElementTree.fromstring(fileContent.read())
            except ParseError as pe:
                log.warn("Can not parse file contents '%s': %r", os.path.basename(srcFile), pe)
                return
        # Extract the relevant fields
        snippet = json.loads(self.komodoToolStr)
        snippet["auto_abbreviation"] = "true"
        snippet["treat_as_ejs"] = "true"
        code = ""
        if xmlContent.find("content") is not None:
            code = xmlContent.find("content").text
        else:
            log.warn("No content to import from snippet.")
            return
        if xmlContent.find("tabTrigger") is not None:
            snippet["name"]  = xmlContent.find("tabTrigger").text
        else:
            log.warn("No trigger to import from snippet.")
            return
        if xmlContent.find("scope") is not None:
            snippet["language"] = self._extractLanguage(xmlContent.find("scope").text)
        if xmlContent.find("description") is not None:
            code = "<% /* " + xmlContent.find("description").text + " */ %>\n" + code
        # Translate code
        code = self._translateSublimeSnippetContent(code)
        snippet["value"] = code.splitlines()
        
        # If there were vars disabled, write it to the file
        if self.disabledVarsInFile:
            self.badFileCount+=1
            relPathToSnippet = ""
            if not self.relativePath == None and not self.relativePath == "..":
                relPathToSnippet = os.path.join(self.relativePath, snippet["name"] + ".komodotool")
            self.addToOutputFile("\n"+relPathToSnippet)

        return snippet
        
    def _extractLanguage(self, scope):
        """Extract language from the scope string from the snippe: eg. 'scope.python' """
        language = scope[scope.index(".")+1:]
        langRegistry = components.classes["@activestate.com/koLanguageRegistryService;1"].getService(components.interfaces.koILanguageRegistryService);
        UnwrapObject(langRegistry)
        langNames = langRegistry.getLanguageNames()
        try:
            return langNames[[name.lower() for name in langNames].index(language.lower())]
        except:
            return ""
    
    # Put all the converted snippets into one place
    def createConvFolder(self, path):
        path = os.path.join(path, self.conversionFolderName)
        if not os.path.exists(path ):
            os.mkdir(path)
        self.conversionFolder = path
        return self.conversionFolder
    
    # It's put into self.convertsionFolder
    # Dumps default text into file from self.unsupportedOrigContent 
    def createUnsupportedFile(self):
        self.unsupportedSnippetsPath = os.path.join(self.conversionFolder, self.unsupportedSnippets)
        try:
            os.remove(self.unsupportedSnippetsPath)
        except OSError:
            # likely means it doesn't exist so skip it.
            pass
        except:
            raise
        self.addToOutputFile(self.unsupportedOrigContent)
        return self.unsupportedSnippetsPath
        
    def addToOutputFile(self, content):
        with open(self.unsupportedSnippetsPath, "a") as badFilesFile:
            badFilesFile.write(content)
    
    # PARAM and $TM_SOFT_TABS variables aren't supported
    # Convert them to something that is obviously not right
    def _disableVarsInTabstops(self, match):
        # Set this var so the file is noted as having an unsupported config
        self.disabledVarsInFile = True
        if not match.group(1) == None:
            return "_"+match.group(1) + "_NOT_SUPPORTED"
        if not match.group(2) == None:
            return "_"+match.group(2) + "_NOT_SUPPORTED"
    
    # Convert tab stops and pull their content along
    # Sublime variables in the default text will be replaced
    # in the next step, 
    def _replSublimeTabstops(self, match):
        if match.group(1):
            return "[[%tabstop"+match.group(1)+":"+match.group(2)+"]]"
        else:
            return "[[%tabstop"+match.group(3)+":"+match.group(3)+"]]"
        
    def _convertSublimeVarsToShortcuts(self, content):
        # Currently this will replace embedded Sublime vars inside
        # tabstops which isn't supported in Komodo.  Things break bad
        subVarsToShortcuts = {
            "$SELECTION" : "require(\"ko/editor\").getColumnNumber()",
            "$TM_CURRENT_LINE" : "require(\"ko/editor\").getLine()",
            "$TM_CURRENT_WORD" : "require(\"ko/editor\").getWord()",
            "$TM_FILENAME" : "require(\"ko/views\").current().get().koDoc.baseName",
            "$TM_FILEPATH" : "require(\"ko/views\").current().get().koDoc.displayPath",
            "$TM_FULLNAME" : "require(\"sdk/system\").env.USERNAME ? require(\"sdk/system\").env.USERNAME : require(\"sdk/system\").env.USER",
            "$TM_LINE_INDEX" : "require(\"ko/editor\").getColumnNumber()",
            "$TM_LINE_NUMBER" : "require(\"ko/editor\").getLineNumber()",
            "$TM_SELECTED_TEXT" : "require(\"ko/editor\").getColumnNumber()",
            "$TM_TAB_SIZE" : "require(\"ko/views\").current().get().koDoc.tabWidth"
            }
        for key, value in subVarsToShortcuts.items():
            content = content.replace(key, "<%= "+value+" %>")
        return content
    
    def _translateSublimeSnippetContent(self, content):
        # Replace the tabstops
        reg = re.compile("\${?(\d+):(.*?)}|\${?(\d+)}?")
        content = reg.sub(self._replSublimeTabstops, content)
        content = self._convertSublimeVarsToShortcuts(content)
        content = re.sub("\$(PARAM\d+)|\$(TM_SOFT_TABS)", self._disableVarsInTabstops, content)#, re.I )not sure if they are case sensitive
        return content
